from metrics_layer.core.exceptions import QueryError
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_generator import MetricsLayerQuery
from metrics_layer.core.sql.query_funnel import FunnelQuery
from metrics_layer.core.sql.query_cumulative_metric import CumulativeMetricsQuery


class SingleSQLQueryResolver:
    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        model=None,
        project=None,
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.has_cumulative_metric = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.explore_name = kwargs.get("explore_name")
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.return_pypika_query = kwargs.get("return_pypika_query")
        self.force_group_by = kwargs.get("force_group_by", False)
        self.project = project
        self.metrics = metrics
        self.dimensions = dimensions
        self.funnel, self.is_funnel_query = self.parse_funnel(funnel)
        self.parse_field_names(where, having, order_by)
        self.model = model
        self.query_type = kwargs.get("query_type")
        if self.query_type is None:
            raise QueryError(
                "Could not determine query_type. Please have connection information for "
                "your warehouse in the configuration or explicitly pass the "
                "'query_type' argument to this function"
            )
        self.parse_input()

    def get_query(self, semicolon: bool = True):
        self.design = MetricsLayerDesign(
            no_group_by=self.no_group_by,
            query_type=self.query_type,
            field_lookup=self.field_lookup,
            model=self.model,
            project=self.project,
        )

        query_definition = {
            "metrics": self.metrics,
            "dimensions": self.dimensions,
            "funnel": self.funnel,
            "where": self.where,
            "having": self.having,
            "order_by": self.order_by,
            "select_raw_sql": self.select_raw_sql,
            "limit": self.limit,
            "return_pypika_query": self.return_pypika_query,
        }
        if self.has_cumulative_metric and self.is_funnel_query:
            raise QueryError("Cumulative metrics cannot be used with funnel queries")

        elif self.has_cumulative_metric:
            query_generator = CumulativeMetricsQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        elif self.is_funnel_query:
            query_generator = FunnelQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        else:
            query_generator = MetricsLayerQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        query = query_generator.get_query(semicolon=semicolon)

        return query

    def get_used_views(self):
        unique_view_names = {f.view.name for f in self.field_lookup.values()}
        return [self.project.get_view(name) for name in unique_view_names]

    def parse_input(self):
        # if self.explore.symmetric_aggregates == "no":
        #     raise NotImplementedError("MetricsLayer does not support turning off symmetric aggregates")

        all_field_names = self.metrics + self.dimensions
        if len(set(all_field_names)) != len(all_field_names):
            # TODO improve this error message
            raise QueryError("Ambiguous field names in the metrics and dimensions")

        for name in self.metrics:
            field = self.get_field_with_error_handling(name, "Metric")
            if field.is_cumulative():
                self.has_cumulative_metric = True
            self.field_lookup[name] = field

        # Dimensions exceptions:
        #   They are coming from a different explore than the metric, not joinable (handled in get_field)
        #   They are not found in the selected explore (handled here)
        # TODO make this better
        metric_view = None if len(self.metrics) == 0 else self.field_lookup[self.metrics[0]].view.name

        for name in self.dimensions:
            field = self.get_field_with_error_handling(name, "Dimension")
            # We will not use a group by if the primary key of the main resulting table is included
            if field.primary_key == "yes" and field.view.name == metric_view and not self.force_group_by:
                self.no_group_by = True
            self.field_lookup[name] = field

        for name in self._where_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Where clause field")

        for name in self._having_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Having clause field")

        for name in self._order_by_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Order by field")

    def get_field_with_error_handling(self, field_name: str, error_prefix: str):
        field = self.project.get_field(field_name, model=self.model)
        if field is None:
            raise QueryError(f"{error_prefix} {field_name} not found")
        return field

    def parse_field_names(self, where, having, order_by):
        self.where = self._check_for_dict(where)
        if self._is_literal(self.where):
            self._where_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.where)
        else:
            self._where_field_names = self.parse_identifiers_from_dicts(self.where)

        self.having = self._check_for_dict(having)
        if self._is_literal(self.having):
            self._having_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.having)
        else:
            self._having_field_names = self.parse_identifiers_from_dicts(self.having)

        self.order_by = self._check_for_dict(order_by)
        if self._is_literal(self.order_by):
            self._order_by_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.order_by)
        else:
            self._order_by_field_names = self.parse_identifiers_from_dicts(self.order_by)

    def parse_funnel(self, funnel: dict):
        if funnel != {}:
            if all(k in funnel for k in ["steps", "within"]):
                is_funnel_query = True
            else:
                raise QueryError("Funnel query must have 'steps' and 'within' keys")
        else:
            is_funnel_query = False
        return funnel, is_funnel_query

    @staticmethod
    def _is_literal(clause):
        return isinstance(clause, str) or clause is None

    @staticmethod
    def parse_identifiers_from_dicts(conditions: list):
        try:
            return [cond["field"] for cond in conditions if "group_by" not in cond]
        except KeyError:
            for cond in conditions:
                if "field" not in cond:
                    break
            raise QueryError(f"Identifier was missing required 'field' key: {cond}")

    @staticmethod
    def _check_for_dict(conditions: list):
        if isinstance(conditions, dict):
            return [conditions]
        return conditions



import os

from metrics_layer.core.model.project import Project
from metrics_layer.core.parse.connections import (
    ConnectionType,
    BaseConnection,
    BigQueryConnection,
    RedshiftConnection,
    SnowflakeConnection,
)

from .github_repo import GithubRepo, LocalRepo
from .manifest import Manifest
from .project_reader_base import ProjectReaderBase
from .project_reader_dbt import dbtProjectReader
from .project_reader_metrics_layer import MetricsLayerProjectReader


class ConfigError(Exception):
    pass


class ProjectLoader:
    def __init__(self, location: str, branch: str = "master", connections: list = [], **kwargs):
        self.kwargs = kwargs
        self.repo = self._get_repo(location, branch, kwargs)
        self._raw_connections = connections
        self._project = None
        self._user = None

    def load(self):
        self._connections = self.load_connections(self._raw_connections)
        self._project = self._load_project()
        return self._project

    @property
    def zenlytic_project(self):
        reader = ProjectReaderBase(repo=self.repo)
        return reader.zenlytic_project if reader.zenlytic_project else {}

    @staticmethod
    def profiles_path():
        home_dir = os.path.expanduser("~")
        profiles_path = os.path.join(home_dir, ".dbt", "profiles.yml")
        return profiles_path

    def get_branch_options(self):
        return self.repo.branch_options

    def _load_project(self):
        self.repo.fetch()
        repo_type = self.repo.get_repo_type()
        if repo_type == "dbt":
            reader = dbtProjectReader(repo=self.repo)
        elif repo_type == "metrics_layer":
            reader = MetricsLayerProjectReader(self.repo)
        else:
            raise TypeError(f"Unknown repo type: {repo_type}, valid values are 'metrics_layer', 'dbt'")

        models, views, dashboards = reader.load()

        self.repo.delete()

        project = Project(
            models=models,
            views=views,
            dashboards=dashboards,
            connection_lookup={c.name: c.type for c in self._connections},
            manifest=Manifest(reader.manifest),
        )
        return project

    def _get_repo(self, location: str, branch: str, kwargs: dict):
        # Config is passed explicitly: this gets first priority
        if location is not None:
            return self._get_repo_from_location(location, branch, kwargs)

        # Next look for environment variables
        repo = self._get_repo_from_environment(kwargs)
        if repo:
            return repo

        raise ConfigError(
            """ We could not find a valid configuration in the environment. Try following the
            documentation (https://docs.zenlytic.com/docs/development_environment/development_environment)
            to properly set your environment variables, or pass the configuration explicitly
        """
        )

    @staticmethod
    def _get_repo_from_location(location: str, branch: str, kwargs: dict):
        if ProjectLoader._is_local(location):
            return LocalRepo(repo_path=location, **kwargs)
        return GithubRepo(repo_url=location, branch=branch, **kwargs)

    @staticmethod
    def _get_repo_from_environment(kwargs: dict):
        prefix = "METRICS_LAYER"
        location = os.getenv(f"{prefix}_LOCATION")
        branch = os.getenv(f"{prefix}_BRANCH", "master")
        repo_type = os.getenv(f"{prefix}_REPO_TYPE")
        if location is None:
            return None

        if ProjectLoader._is_local(location):
            return LocalRepo(repo_path=location, repo_type=repo_type, **kwargs)
        return GithubRepo(repo_url=location, branch=branch, repo_type=repo_type, **kwargs)

    @staticmethod
    def _is_local(location: str):
        return "https://" not in location

    @staticmethod
    def load_connections(connections: list):
        class_lookup = {
            ConnectionType.snowflake: SnowflakeConnection,
            ConnectionType.bigquery: BigQueryConnection,
            ConnectionType.redshift: RedshiftConnection,
        }
        results = []
        for connection in connections:
            if isinstance(connection, BaseConnection):
                connection_class = connection
            else:
                connection_type = connection["type"].upper()
                connection_class = class_lookup[connection_type](**connection)
            results.append(connection_class)
        return results

    def get_connections_from_profile(profile_name: str, target: str = None):
        profile_path = ProjectLoader.profiles_path()
        profiles_directory = os.path.dirname(profile_path)
        profiles_dict = dbtProjectReader.read_yaml_if_exists(profile_path)
        if profiles_dict is None:
            raise ConfigError(f"Could not find dbt profiles.yml at {profile_path}")

        profile = profiles_dict.get(profile_name)

        if profile is None:
            raise ConfigError(f"Could not find profile {profile_name} in profiles.yml at {profile_path}")

        if target is None:
            target = profile["target"]

        target_dict = profile["outputs"].get(target)

        if target_dict is None:
            raise ConfigError(
                f"Could not find target {target} in profile {profile_name} in profiles.yml at {profile_path}"
            )

        return [{**target_dict, "directory": profiles_directory, "name": profile_name}]



from .base import MetricsLayerBase
from metrics_layer.core.exceptions import QueryError


class AccessGrant(MetricsLayerBase):
    def __init__(self, definition: dict = {}) -> None:

        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "user_attribute", "allowed_values"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Access Grant missing required key {k}")


class Model(MetricsLayerBase):
    def __init__(self, definition: dict = {}, project=None) -> None:

        self.project = project
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "connection"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Model missing required key {k}")

    @property
    def access_grants(self):
        return self._definition.get("access_grants", [])

    def collect_errors(self):
        if not self.valid_name(self.name):
            return [self.name_error("model", self.name)]
        return []

    def printable_attributes(self):
        to_print = ["name", "type", "label", "group_label", "connection"]
        attributes = self.to_dict()
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    def get_mappings(self):
        if self.mappings is None:
            return {}

        dimension_mapping = {}
        for _, mapped_values in self.mappings.items():

            for mapped_from_field in mapped_values:
                from_field = self.project.get_field(mapped_from_field)
                from_join_hash = self.project.join_graph.join_graph_hash(from_field.view.name)
                if from_field.field_type in {"dimension_group", "measure"}:
                    raise QueryError(
                        "This mapping is invalid because it contains a dimension group or "
                        f"a measure. Mappings can only contain dimensions. Mapping with {from_field.id()}"
                    )

                for mapped_to_field in mapped_values:
                    if mapped_to_field != mapped_from_field:
                        to_field = self.project.get_field(mapped_to_field)
                        to_join_hash = self.project.join_graph.join_graph_hash(to_field.view.name)
                        map_data = {
                            "field": mapped_to_field,
                            "from_join_hash": from_join_hash,
                            "to_join_hash": to_join_hash,
                        }
                        dimension_mapping[mapped_from_field] = map_data

        return dimension_mapping
